"""Predicates for DIF v1.3."""

from rdflib import DCTERMS, OWL, SDO, URIRef

from ..namespace import (
    EDUCOR,
    FABIO,
    bibframe_lite_relation,
    citedcat,
    fabio,
    metadata4ing,
    modalia,
    rec,
)

EDUCATIONAL_RESOURCE_CLASS = EDUCOR["EducationalResource"]
ORGANIZATION_CLASS = SDO.Organization
PERSON_CLASS = SDO.Person

DISCIPLINE_PREDICATE = FABIO["hasDiscipline"]
SUPPORTING_COMMUNITY_PRED = bibframe_lite_relation.supportinghost
RECOMMENDING_COMMUNITY_PRED = rec.recommender

MEDIA_TYPES_PREDICATE = modalia.hasMediaType
PROFICIENCY_LEVEL_PREDICATE = modalia.requiresProficiencyLevel
FILE_SIZE_PREDICATE = SDO.fileSize
TARGET_GROUP_PREDICATE = modalia.hasTargetGroup
TITLE_PREDICATE = DCTERMS.title
LINK_PREDICATE = SDO.url
VERSION_PREDICATE = SDO.version
SUBTITLE_PREDICATE = fabio.hasSubtitle
LICENSE_PREDICATE = DCTERMS.license
LANGUAGE_PREDICATE = DCTERMS.language
DATE_PUBLISHED_PREDICATE = SDO.datePublished
LEARNING_RESOURCE_TYPE_PREDICATE = modalia.hasLearningType
KEYWORDS_PREDICATE = SDO.keywords
FILE_FORMAT_PREDICATE = DCTERMS.format
DESCRIPTION_PREDICATE = DCTERMS.description

AUTHOR_PREDICATE = SDO.author
AUTHOR_UNORDERED_PREDICATE = URIRef("https://dalia.education/authorUnordered")

ORGANIZATION_NAME_PREDICATE = SDO.name
ORGANIZATION_WIKIDATA_PREDICATE = OWL.sameAs
ORGANIZATION_ROR_PREDICATE = metadata4ing.hasRorId

PERSON_FAMILY_NAME_PREDICATE = SDO.familyName
PERSON_GIVEN_NAME_PREDICATE = SDO.givenName
PERSON_ORCID_PREDICATE = metadata4ing.orcidId

IS_SUPPLEMENTED_BY_PREDICATE = citedcat.isSupplementedBy
NEW_VERSION_OF_PREDICATE = modalia.isNewVersionOf



"""Constants for DIF v1.3."""

import click

#: A single UUID to represent the resource under the ``dalia.oer`` prefix. Generate this UUID
#: yourself.
DIF_HEADER_ID = "DALIA_ID"
#: The SPDX license describing the upstream resource's terms and conditions. Dashes, spaces,
#: and capitalization are normalized.
#:
#: Warning: don't confuse this with the licensing of the metadata curated in the DIF, which
#: is licensed independently (we suggest CC0 for maximum reusability)
#:
#: This field also accepts the literal strings
#: ``proprietary`` or ``unlicensed``. For example, videos on YouTube are often licensed with
#: YouTube's proprietary license if the creator did not opt into using CC-BY-4.0.
DIF_HEADER_LICENSE = "License"
#: The URL link to the OER. Ideally, this is a DOI or other persistent identifier, but it can
#: be any URL that resolves
DIF_HEADER_LINK = "Link"
#: The language of the OER, written as an ISO two-letter code (e.g., `en` for English, `de`
#: for German)
DIF_HEADER_LANGUAGE = "Language"
#: The title of the resource. Please write this in the same language mentioned in the "Language"
#: column. If a colon `:` is present, the first one will be used to split the title into a title
#: and subtitle, which get put in different fields in DIF v1.3.
DIF_HEADER_TITLE = "Title"
#: A textual description of the resource. Please write this in the same language mentioned in the
#: "Language" column
DIF_HEADER_DESCRIPTION = "Description"
#: A list of authors, separated by asterisks. Each author should be written with family names, then
#: a comma, then given names. If an ORCID is available, then it can be added with a colon ``:``
#: then inside curly braces like in the following example.
#:
#: For example, `Kremer, Dominik : {https://orcid.org/0000-0003-1244-7363} * Geiger, Jonathan : {https://orcid.org/0000-0002-0452-7075}`
DIF_HEADER_AUTHORS = "Authors"
#: A list of UUIDs corresponding to pre-curated communities in DALIA. If you would like to request
#: a new one, please email charles.hoyt@ac.rwth-aachen.de.
DIF_HEADER_COMMUNITY = "Community"
#: The disciplines (e.g., chemistry, biology) covered by the resource encoded using
#: URIs from the DINI-KIM Hochschulfächersystematik resource, like
#: https://w3id.org/kim/hochschulfaechersystematik/n7.
#:
#: See https://w3id.org/kim/hcrt/scheme.html where you can search for the term and copy
#: the correct URI. When a discipline is missing from the resource, please either directly make an
#: issue on the upstream [GitHub repository](https://github.com/dini-ag-kim/hcrt) or email
#: charles.hoyt@ac.rwth-aachen.de for help
DIF_HEADER_DISCIPLINE = "Discipline"
#: A list of file extensions used by the resource (e.g., `.pdf`).
#: If multiple are available, please use an asterisk as a delimiter (e.g., `.pdf * .zip`).
DIF_HEADER_FILE_FORMAT = "FileFormat"
#: Free text keywords for the resource, delimited by an asterick (e.g.,
#: `digital humanities * culture`). Please write keywords in the same language
#: mentioned in the "Language" column
DIF_HEADER_KEYWORDS = "Keywords"
#: The learning resource type. Choose one of the keys in
#: :data:`dalia_dif.dif13.picklist.LEARNING_RESOURCE_TYPES`.
DIF_HEADER_LEARNING_RESOURCE_TYPE = "LearningResourceType"
#: The media type says what modality the learning resource has (e.g., audio, video, text).
#: This is a more broad category than learning resource type. Choose from
#: one of the keys in :data:`dalia_dif.dif13.picklist.MEDIA_TYPES`.
DIF_HEADER_MEDIA_TYPE = "MediaType"
#: One of the five proficiency levels described by the Dreyfus adaptive learning model
#: (https://doi.org/10.1177/02704676042649): novice, advanced beginner, competent, proficient, or
#: expert. These are encoded in :data:`dalia_dif.dif13.picklist.PROFICIENCY_LEVELS`
DIF_HEADER_PROFICIENCY_LEVEL = "ProficiencyLevel"
#: The date of publication, written in ISO standard format YYYY-MM-DD
DIF_HEADER_PUBLICATION_DATE = "PublicationDate"
#: The target audience (e.g., school students, bachelor's level students, data stewards, etc.).
#: Choose from the keys in :data:`dalia_dif.dif13.picklist.TARGET_GROUPS`.
DIF_HEADER_TARGET_GROUP = "TargetGroup"
#: Links to related works. These are curated as an asterisk-delimited list of predicate-value pairs
#: like: `isPartOf:https://doi.org/10.11588/heidicon/1738716` where the first part of the string is
#: a key from :data:`dalia_dif.dif13.picklist.RELATED_WORKS_RELATIONS` followed by a colon, then
#: the URL (ideally a DOI or other persistent identifier) to the target of the relation.
DIF_HEADER_RELATED_WORK = "RelatedWork"
#: The size, in megabytes (MB) of the file (e.g., `0.142`). Use a maximum of 3 places after the
#: decimal point. Do not write `MB`.
DIF_HEADER_SIZE = "Size"
#: The version of the OER. The same OER that is updated over time might have different versions,
#: each of which could be assigned their own record (and UUID) in DALIA.
DIF_HEADER_VERSION = "Version"

#: separator used for list fields
DIF_SEPARATOR = " * "

REQUIRED = {
    DIF_HEADER_ID,
    DIF_HEADER_LINK,
    DIF_HEADER_TITLE,
    DIF_HEADER_LANGUAGE,
    DIF_HEADER_LICENSE,
}


@click.command()
def main() -> None:
    """Create a curation guide based on this document."""
    import sys
    from pathlib import Path
    from textwrap import dedent

    from sphinx.pycode import ModuleAnalyzer

    module_root = Path(__file__).parent.parent.resolve()
    repo_root = module_root.parent.parent.resolve()

    path = repo_root.joinpath("docs", "curation.md")

    analyzer = ModuleAnalyzer.for_file(__file__, "mymodule")
    analyzer.analyze()

    text = dedent("""\
    # DALIA DIF v1.3 CSV Curation Guide

    This guide contains an explanation of the headers that can
    appear within a DIF v1.3 CSV file. Each has an explanation of what
    data type goes in each (string, URI reference, date, etc.), whether
    it's required or optional, and an explanation of how to curate values.

    ## Columns
    """)

    # Get the extracted documentation
    for (_, variable_name), docs_lines in analyzer.attr_docs.items():
        if variable_name.startswith("DIF_HEADER"):
            variable_value = getattr(sys.modules[__name__], variable_name)
            docs = "\n".join(docs_lines)
            required_text = " (required)" if variable_value in REQUIRED else ""
            column_text = f"""
### `{variable_value}`{required_text}

{docs}"""
            text += column_text

    text = text.rstrip()
    text += dedent(f"""

    ## Colophon

    Additional example files can be found in
    https://github.com/data-literacy-alliance/dalia-curation.

    This guide was autogenerated from the code documentation
    in `{Path(__file__).relative_to(repo_root)}`
    """)

    path.write_text(text)


if __name__ == "__main__":
    main()



"""A legacy implementation of author parsing DIF v1.3 from Frank Lange."""

import re

import base32_crockford
from rdflib import RDF, Graph
from rdflib.term import BNode, Literal, Node, URIRef

from ..constants import DIF_SEPARATOR
from ..predicates import (
    AUTHOR_PREDICATE,
    AUTHOR_UNORDERED_PREDICATE,
    ORGANIZATION_CLASS,
    ORGANIZATION_NAME_PREDICATE,
    ORGANIZATION_ROR_PREDICATE,
    ORGANIZATION_WIKIDATA_PREDICATE,
    PERSON_CLASS,
    PERSON_FAMILY_NAME_PREDICATE,
    PERSON_GIVEN_NAME_PREDICATE,
    PERSON_ORCID_PREDICATE,
)
from ..utils import create_rdf_collection

#: part of this regex is from https://ror.readme.io/docs/identifier
ROR_RE = re.compile(r"^https://ror\.org/0[a-hj-km-np-tv-z|0-9]{6}\d{2}$")
ORCID_RE = re.compile(r"^https?://orcid\.org/\d{4}-\d{4}-\d{4}-\d{3}[\dX]$")
WIKIDATA_RE = re.compile(r"^http://www\.wikidata\.org/entity/Q\d*$")


def _create_organization_author(g: Graph, organization: str) -> Node:
    match = re.search(r"^(?P<name>.+)\s:\s{organization\s?(?P<identifier>.*)}$", organization)

    if not match:
        raise Exception(f'Could not match regex for organization "{organization}"')

    organization_name = match.group("name").strip()
    organization_identifier = match.group("identifier")

    organization_node = BNode()
    g.add((organization_node, RDF.type, ORGANIZATION_CLASS))
    g.add((organization_node, ORGANIZATION_NAME_PREDICATE, Literal(organization_name)))

    if organization_identifier and (organization_identifier := organization_identifier.strip()):
        if is_valid_ror_id(organization_identifier):
            g.add((organization_node, ORGANIZATION_ROR_PREDICATE, Literal(organization_identifier)))
        elif is_valid_wikidata_concept_uri(organization_identifier):
            g.add(
                (
                    organization_node,
                    ORGANIZATION_WIKIDATA_PREDICATE,
                    URIRef(organization_identifier),
                )
            )
        else:
            raise Exception(f'Invalid identifier in "{organization}": {organization_identifier}')

    return organization_node


def _create_person_author(g: Graph, person: str) -> Node:
    person_substrings = person.split(" : ")
    name_substrings = person_substrings[0].split(",")

    person_node = BNode()
    g.add((person_node, RDF.type, PERSON_CLASS))
    g.add((person_node, PERSON_FAMILY_NAME_PREDICATE, Literal(name_substrings[0].strip())))

    if len(name_substrings) > 1:
        g.add((person_node, PERSON_GIVEN_NAME_PREDICATE, Literal(name_substrings[1].strip())))

    if (len(person_substrings) > 1) and (
        match := re.match(r"^{(?P<identifier>.*)}$", person_substrings[1])
    ):
        person_identifier = match.group("identifier").strip()
        if is_valid_orcid(person_identifier):
            g.add((person_node, PERSON_ORCID_PREDICATE, Literal(person_identifier)))
        else:
            raise Exception(f'Invalid identifier in "{person}": {person_identifier}')

    return person_node


def _create_author(g: Graph, author: str) -> Node | None:
    if not author:
        return None

    if re.search(r"\s:\s{organization.*}", author):
        return _create_organization_author(g, author)
    return _create_person_author(g, author)


def _add_ordered_list_of_authors_to_lr(g: Graph, lr_node: Node, authors: str) -> None:
    author_nodes = []
    for author in authors.split(DIF_SEPARATOR):
        if author_node := _create_author(g, author.strip()):
            author_nodes.append(author_node)

    if authors_list := create_rdf_collection(g, author_nodes):
        g.add((lr_node, AUTHOR_PREDICATE, authors_list))


def _add_unordered_set_of_authors_to_lr(g: Graph, lr_node: Node, authors: str) -> None:
    for author in authors.split(DIF_SEPARATOR):
        if author_node := _create_author(g, author.strip()):
            g.add((lr_node, AUTHOR_UNORDERED_PREDICATE, author_node))


def add_authors_to_lr(g: Graph, lr_node: Node, authors: str) -> None:
    if not authors.strip():
        raise Exception("Empty Authors field")

    if authors.strip().lower() == "n/a":
        return None

    _add_ordered_list_of_authors_to_lr(g, lr_node, authors)
    _add_unordered_set_of_authors_to_lr(g, lr_node, authors)
    return None


# see https://support.orcid.org/hc/en-us/articles/360006897674-Structure-of-the-ORCID-Identifier#2-checksum
def _generate_check_digit(base_digits: str) -> str:
    total = 0
    for c in base_digits:
        digit = int(c)
        total = (total + digit) * 2
    remainder = total % 11
    result = (12 - remainder) % 11
    return "X" if result == 10 else str(result)


def _check_orcid_checksum(orcid: str) -> bool:
    digits = orcid.split("/")[-1].replace("-", "")
    return _generate_check_digit(digits[:-1]) == digits[-1]


def is_valid_orcid(orcid: str) -> bool:
    """Check if an ORCID string is syntactically correct. Does not check for existence of the ORCID."""
    return bool(ORCID_RE.search(orcid)) and _check_orcid_checksum(orcid)


def _check_ror_id_checksum(ror_id: str) -> bool:
    # inverse of https://github.com/ror-community/ror-api/blob/034592be42654ad6638103e2a7638cab344d2880/rorapi/management/commands/generaterorid.py#L6
    record = ror_id.split("/")[-1]
    checksum = record[-2:]
    n_encoded = record[0:-2]
    n = base32_crockford.decode(n_encoded)

    return checksum == str(98 - ((n * 100) % 97)).zfill(2)


def is_valid_ror_id(ror_id: str) -> bool:
    """Check if a ROR ID is syntactically correct. Does not check for existence of the ROR ID."""
    return bool(ROR_RE.search(ror_id)) and _check_ror_id_checksum(ror_id)


def is_valid_wikidata_concept_uri(concept_uri: str) -> bool:
    """Check if a Wikidata concept URI is syntactically correct.

    Does not check for existence of the Wikidata Entity behind this URI.
    """
    return bool(WIKIDATA_RE.search(concept_uri))



"""Terms from The Recommendation Ontology (http://purl.org/ontology/rec/core#)."""

from rdflib import Namespace

__all__ = [
    "REC",
]

REC = Namespace("http://purl.org/ontology/rec/core#")

# Properties
recommender = REC["recommender"]



"""Picklists for DIF v1.3."""

from rdflib import PROV, SDO, URIRef

from .predicates import (
    IS_SUPPLEMENTED_BY_PREDICATE,
    NEW_VERSION_OF_PREDICATE,
    RECOMMENDING_COMMUNITY_PRED,
    SUPPORTING_COMMUNITY_PRED,
)
from ..namespace import bibo, citedcat, hcrt, modalia

__all__ = [
    "COMMUNITY_RELATIONS",
    "LEARNING_RESOURCE_TYPES",
    "MEDIA_TYPES",
    "PROFICIENCY_LEVELS",
    "PROPRIETARY_LICENSE",
    "RELATED_WORKS_RELATIONS",
    "TARGET_GROUPS",
]

TARGET_GROUPS = {
    "student (school)": modalia.StudentSchool,
    "student (ba)": modalia.BachelorStudent,
    "student (ma)": modalia.MastersStudent,
    "student (phd)": modalia.PhDStudent,
    "data steward": modalia.DataSteward,
    "teacher (school)": modalia.TeacherSchool,
    "teacher (higher education)": modalia.TeacherHighEducation,
    "researcher": modalia.Researcher,
    "content provider": modalia.ContentProvider,
}
MEDIA_TYPES = {
    "audio": SDO.AudioObject,
    "video": SDO.VideoObject,
    "text": SDO.Text,
    "presentation": SDO.PresentationDigitalDocument,
    "code": modalia.Code,  # FIXME why not SDO.SoftwareSourceCode
    "image": SDO.ImageObject,
    "multipart": modalia.Multipart,
}
PROFICIENCY_LEVELS = {
    "novice": modalia.Novice,
    "advanced beginner": modalia.Beginner,
    "competent": modalia.Competent,
    "proficient": modalia.Proficient,
    "expert": modalia.Expert,
}
RELATED_WORKS_RELATIONS = {
    "isPartOf": modalia.isPartOf,
    "hasPart": SDO.hasPart,
    "isBasedOn": modalia.isBasedOn,
    "isNewerVersionOf": NEW_VERSION_OF_PREDICATE,
    "isSupplementOf": citedcat.isSupplementTo,
    "isSupplementTo": citedcat.isSupplementTo,
    "isSupplementedBy": IS_SUPPLEMENTED_BY_PREDICATE,
    "wasRevisionOf": PROV.wasRevisionOf,
    "isDuplicateOf": modalia.isDuplicateOf,
    "isTranslationOf": modalia.isTranslationOf,
}


# Entries that will be ignored during the mapping because they shall be used as media type.
# Format: URI -> media type to be used instead
MEDIA_TYPE_EXCEPTIONS = {
    hcrt.audio: "audio",
    hcrt.image: "image",
    hcrt.slide: "presentation",
    hcrt.text: "text",
    hcrt.video: "video",
}

LEARNING_RESOURCE_TYPES: dict[str, URIRef] = {
    "educational game": hcrt.educational_game,
    "case study": hcrt.case_study,
    "experiment": hcrt.experiment,
    "diagram": hcrt.diagram,
    "course": hcrt.course,
    "hcrt:diagram": hcrt.diagram,
    "hcrt:educational_game": hcrt.educational_game,
    "hcrt:case_study": hcrt.case_study,
    "hcrt:experiment": hcrt.experiment,
    # SDO
    "podcastseries": SDO.PodcastSeries,
    "schema:podcastseries": SDO.PodcastSeries,
    "code": SDO.SoftwareSourceCode,
    # BIBO
    "article": bibo.Article,
    "book": bibo.Book,
    "report": bibo.Report,
    "webpage": bibo.Webpage,
    "thesis": bibo.Thesis,
    "bibo:article": bibo.Article,
    "bibo:book": bibo.Book,
    "bibo:report": bibo.Report,
    "bibo:webpage": bibo.Webpage,
    "bibo:thesis": bibo.Thesis,
    # Internal
    "poster": modalia.Poster,
    "lecture": modalia.Lecture,
    "tutorial": modalia.Tutorial,
    "codenotebook": modalia.CodeNotebook,
    "bestpractices": modalia.BestPractices,
    "cookbook": modalia.Cookbook,
    "mo:poster": modalia.Poster,
    "mo:lecture": modalia.Lecture,
    "mo:tutorial": modalia.Tutorial,
    "mo:codenotebook": modalia.CodeNotebook,
    "mo:best-practices": modalia.BestPractices,
    "mo:cookbook": modalia.Cookbook,
    # added by charlie to clean up errors
    "interview": modalia.Interview,
    # Mappings provided by Petra and Abdel for MVP1
    # "Video": None,
    # "Presentation": None,
    # "Website": bibo.Webpage,
    # "Case studies and tutorials": hcrt.case_study,
    # "Video (Serie)": None,
    # "Conference proceeding": None,
    # "free self-paced course": None,
    # "Blogpost": None,
    # "Conference paper": bibo.Article,
    # "Figure": hcrt.diagram,
    # "Git repository": None,
    # "Jupyter Notebook": modalia.JupyterNotebook,
    # "Project deliverable": bibo.Report,
    # "Project milestone": bibo.Report,
    # "Proposal": None,
    # "Working paper": bibo.Article,
    # "Best practices and guidelines": modalia.BestPractices,
    # "Curriculum": None,
    # "Guide": None,
    # "Handbook": bibo.Book,
    # "Image": None,
    # "in person workshop": modalia.Workshop,
    # "Journal article": bibo.Article,
    # "Lesson (Self-Study Unit)": None,
    # "Other": None,
    # "Preprint": bibo.Article,
    # "School": None,
    # "Software": None,
    # "Workflow": None,
    # "Youtube Channel": None,
}

COMMUNITY_RELATIONS = {
    "S": SUPPORTING_COMMUNITY_PRED,
    "R": RECOMMENDING_COMMUNITY_PRED,
}

PROPRIETARY_LICENSE = modalia.ProprietaryLicense



"""Full-text indexing utilities.

This module contains the functions for generating a full text index in SQLite from open
educational resources encoded in RDF using the DIF v1.3 schema.

.. code-block:: python

    from dalia_ingest.model import dif13_to_sqlite_fti, query_sqlite_fti

    # give a path or list of paths to TTL files to parse
    ttl_path = ...

    # constructs an in-memory database, should be done once
    # during startup of the backend
    conn = dif13_to_sqlite_fti(ttl_path)

    # many queries can be made over the same in-memory,
    # database object. ``*`` can be used as a wildcard at
    # the end (but not the beginning) of the string
    uuids = query_sqlite_fti("chem*", conn)
"""

import sqlite3
from collections import defaultdict
from contextlib import closing
from pathlib import Path
from textwrap import dedent

import pandas as pd
import rdflib

from dalia_dif.namespace import DALIA_OER

__all__ = [
    "dif13_to_sqlite_fti",
    "query_sqlite_fti",
    "write_sqlite_fti",
]


def query_sqlite_fti(query: str, db: str | Path | sqlite3.Connection) -> list[str]:
    """Get UUIDs for documents matching the query.

    :param query: The query string, like `chem`. Can also include wildcards like in
        `chem*`.
    :param db: Either a path to a SQLite database file or an already-established
        connection

    :returns: A list of UUIDs for OERs that match the query
    """
    sql = _get_fts_sql(query)

    if isinstance(db, str | Path):
        db = Path(db).expanduser().resolve()
        with closing(sqlite3.connect(db.as_posix())) as conn, closing(conn.cursor()) as cursor:
            results = cursor.execute(sql).fetchall()
    elif isinstance(db, sqlite3.Connection):
        with closing(db.cursor()) as cursor:
            results = cursor.execute(sql).fetchall()
    else:
        raise TypeError

    # unpack UUIDs and throw away scores
    return [uuid for uuid, _score in results]


def _get_fts_sql(query: str) -> str:
    # Test FTS query (e.g., search all fields for "python") note that the bm25 weights
    return dedent(f"""\
        SELECT uuid, bm25(documents, 0.0, 5.0, 1.0, 0.5)
        FROM documents
        WHERE documents MATCH '{query}'
        ORDER BY rank;
    """)  # noqa:S608


def dif13_to_sqlite_fti(paths: str | Path | list[str | Path]) -> sqlite3.Connection:
    """Construct an in-memory SQLite database with a full-text index over OERs encoded in DIF v1.3.

    :param paths: The path or paths to turtle files encoding OERs in DIF v1.3

    :returns: An in-memory SQLite database object that can be queried

    Example usage:

    .. code-block:: python

        ttl_path = ...
        conn = dif13_to_sqlite_fti(ttl_path)
        uuids = query_sqlite_fti("chem*", conn)
    """
    graph = rdflib.Graph()

    if isinstance(paths, str | Path):
        graph.parse(paths)
    elif isinstance(paths, list):
        for path in paths:
            graph.parse(path)
    else:
        raise TypeError(f"`paths` should be a path or list of paths. Got: ({type(paths)}) {paths}")

    return graph_to_conn(graph)


def graph_to_conn(graph: rdflib.Graph) -> sqlite3.Connection:
    df = graph_to_df(graph)
    conn = sqlite3.connect(":memory:")
    _dif13_df_to_sqlite(df, conn)
    return conn


def graph_to_df(graph: rdflib.Graph) -> pd.DataFrame:
    titles = defaultdict(set)
    descriptions = defaultdict(set)
    keywords = defaultdict(set)
    for uri, title, description, keyword in list(  # type:ignore[misc]
        graph.query(DIF13_DISCIPLINE_PREDICATE_SPARQL_QUERY)
    ):
        uuid = uri.removeprefix(str(DALIA_OER))
        if title:
            titles[uuid].add(title)
        if description:
            descriptions[uuid].add(description)
        if keyword:
            keywords[uuid].add(keyword)

    rows = [
        (
            uuid,
            " ".join(titles.get(uuid, [])),
            " ".join(descriptions.get(uuid, [])),
            " ".join(keywords.get(uuid, [])),
        )
        for uuid in set(titles).union(descriptions).union(keywords)
    ]
    df = pd.DataFrame(rows, columns=["uuid", "title", "description", "keywords"])
    return df


#: Query RDF encoded in DIF v1.3 for an OER's identifier, title, description, and keywords
DIF13_DISCIPLINE_PREDICATE_SPARQL_QUERY = """\
SELECT
    ?oer
    ?title
    ?description
    ?keyword
WHERE {
    ?oer a educor:EducationalResource .
    OPTIONAL { ?oer dcterms:title ?title . }
    OPTIONAL { ?oer dcterms:description ?description . }
    OPTIONAL { ?oer schema:keywords ?keyword . }
}
"""


def _dif13_df_to_sqlite(df: pd.DataFrame, conn: sqlite3.Connection) -> None:
    """Write a dataframe to a SQLite database (which could be in-memory)."""
    # Enable FTS5 extension (usually built-in with modern SQLite)
    # Create FTS5 virtual table
    query = dedent("""\
        CREATE VIRTUAL TABLE documents USING fts5(
            uuid,
            title,
            description,
            keywords,
            prefix='2 3 4 5 6',
            tokenize = 'porter'
        )
    """)
    with closing(conn.cursor()) as cursor:
        cursor.execute(query)
    df.to_sql("documents", conn, if_exists="append", index=False)


def write_sqlite_fti(graph: rdflib.Graph, path: Path) -> None:
    """Write a SQLite database with a full text index."""
    df = graph_to_df(graph)
    with closing(sqlite3.connect(path.as_posix())) as conn:
        _dif13_df_to_sqlite(df, conn)



"""Reader for DIF v1.3."""

import re
from collections import Counter
from pathlib import Path
from typing import TextIO

import click
import rdflib
from pydantic_extra_types.language_code import ISO639_3
from pydantic_metamodel.api import PredicateObject, RDFResource, Year
from pystow.utils import safe_open_dict_reader
from rdflib import URIRef
from tqdm import tqdm

from .community import LOOKUP_DICT_COMMUNITIES
from .model import (
    AuthorDIF13,
    EducationalResourceDIF13,
    OrganizationDIF13,
)
from .picklists import (
    LEARNING_RESOURCE_TYPES,
    MEDIA_TYPES,
    PROFICIENCY_LEVELS,
    PROPRIETARY_LICENSE,
    RELATED_WORKS_RELATIONS,
    TARGET_GROUPS,
)
from ..namespace import DALIA_COMMUNITY, SPDX_LICENSE, bind, get_base_graph
from ..utils import cleanup_languages

__all__ = [
    "parse_dif13_row",
    "read_dif13",
    "read_dif13_into_rdflib",
    "write_dif13_jsonl",
    "write_dif13_rdf",
]

DELIMITER = " * "

ORCID_RE = re.compile(r"^\d{4}-\d{4}-\d{4}-\d{3}(\d|X)$")

ORCID_URI_PREFIX = "https://orcid.org/"
ROR_URI_PREFIX = "https://ror.org/"
WIKIDATA_URI_PREFIX = "http://www.wikidata.org/entity/"
COMMUNITY_RELATION_RE = re.compile(r"^(?P<name>.*)\s\((?P<relation>S|R|SR|RS)\)$")

#: Keep track of fields in DIF CSV files that haven't
#: been explicitly processed
UNPROCESSED: Counter[str] = Counter()


def write_dif13_rdf(
    oers: EducationalResourceDIF13 | list[EducationalResourceDIF13],
    *,
    path: Path | None = None,
    format: str | None = None,
) -> None:
    """Write OERs as DIF v1.3 RDF."""
    if isinstance(oers, EducationalResourceDIF13):
        oers = [oers]
    graph = rdflib.Graph()
    bind(graph)
    for er in oers:
        graph += er.get_graph()
    if format is None:
        format = "turtle"
    if path is None:
        click.echo(graph.serialize(format=format))
    else:
        graph.serialize(path, format=format)


def write_dif13_jsonl(
    oers: EducationalResourceDIF13 | list[EducationalResourceDIF13], *, path: Path | None = None
) -> None:
    """Write OERs as DIF v1.3 JSON lines."""
    if isinstance(oers, EducationalResourceDIF13):
        oers = [oers]
    lines = (o.model_dump_json(exclude_none=True, exclude_defaults=True) for o in oers)
    if path is None:
        for line in lines:
            click.echo(line)
    else:
        with path.open("w") as file:
            file.write("\n".join(lines))


def read_dif13(
    path: str | Path | TextIO, *, error_accumulator: list[str] | None = None
) -> list[EducationalResourceDIF13]:
    """Parse DALIA records."""
    if isinstance(path, str) and (path.startswith("http://") or path.startswith("https://")):
        from io import StringIO

        import requests

        with requests.get(path, timeout=5) as res:
            sio = StringIO(res.text)
            sio.name = path.split("/")[-1]
            return read_dif13(sio)

    if isinstance(path, (str, Path)):
        file_name = Path(path).name
    else:
        file_name = path.name
    with safe_open_dict_reader(path, delimiter=",") as reader:
        return [
            oer
            for idx, record in enumerate(reader, start=2)
            if (oer := parse_dif13_row(file_name, idx, record, error_accumulator=error_accumulator))
            is not None
        ]


def read_dif13_into_rdflib(
    path: str | Path | TextIO, *, error_accumulator: list[str] | None = None
) -> rdflib.Graph:
    """Read DALIA DIF v1.3 records into a RDFlib graph."""
    graph = get_base_graph()
    for record in read_dif13(path, error_accumulator=error_accumulator):
        record.add_to_graph(graph)
    return graph


def parse_dif13_row(
    file_name: str,
    idx: int,
    row: dict[str, str],
    *,
    future: bool = False,
    error_accumulator: list[str] | None = None,
) -> EducationalResourceDIF13 | None:
    """Convert a row in a DALIA curation file to a resource, or return none if unable."""
    supporting_communities, recommending_communities = _process_communities(
        file_name, idx, row, error_accumulator=error_accumulator
    )

    external_uris = _pop_split(row, "Link")
    if future and (n4c_id := row.pop("N4C_ID", None)):
        external_uris.append(n4c_id)

    title, _, subtitle = map(str.strip, row.pop("Title").partition(":"))

    uuid = row.pop("DALIA_ID", None) or row.pop("uuid", None)
    if not uuid:
        _log(file_name, idx, "no UUID given", error_accumulator=error_accumulator)
        return None

    try:
        rv = EducationalResourceDIF13(
            uuid=uuid,
            title=title,
            subtitle=subtitle or None,
            authors=_process_authors(file_name, idx, row, error_accumulator=error_accumulator),
            license=_process_license(row),
            links=external_uris,
            supporting_communities=supporting_communities,
            recommending_communities=recommending_communities,
            description=row.pop("Description").strip() or None,
            disciplines=_process_disciplines(
                file_name, idx, row, error_accumulator=error_accumulator
            ),
            file_formats=_process_formats(row),
            keywords=_pop_split(row, "Keywords"),
            languages=_process_languages(row),
            learning_resource_types=_process_learning_resource_types(
                file_name, idx, row, error_accumulator=error_accumulator
            ),
            media_types=_process_media_types(
                file_name, idx, row, error_accumulator=error_accumulator
            ),
            proficiency_levels=_process_proficiency_levels(row),
            publication_date=_process_publication_date(row),
            target_groups=_process_target_groups(
                file_name, idx, row, error_accumulator=error_accumulator
            ),
            related_works=_process_related_works(
                file_name, idx, row, error_accumulator=error_accumulator
            ),
            file_size=_process_size(row),
            version=row.pop("Version") or None,
        )
    except ValueError as e:
        _log(
            getattr(file_name, "name", ""),
            idx,
            str(e),
            error_accumulator=error_accumulator,
        )
        return None
    for k, v in row.items():
        if v and v.strip():
            UNPROCESSED[k] += 1
    return rv


def _pop_split(d: dict[str, str], key: str) -> list[str]:
    s = d.pop(key, None)
    if not s:
        return []
    return [y for x in s.split(DELIMITER) if (y := x.strip())]


def _log(
    file_name: str, line: int, text: str, *, error_accumulator: list[str] | None = None
) -> None:
    msg = f"[{file_name} line:{line}] {text}"
    if error_accumulator is None:
        tqdm.write(msg)
    else:
        error_accumulator.append(msg)


def _process_publication_date(row: dict[str, str]) -> Year | str | None:
    date: str | None = row.pop("PublicationDate", None)
    if not date:
        return None
    try:
        year = int(date)
    except ValueError:
        return date
    else:
        return Year(year)


def _process_disciplines(
    file_name: str,
    idx: int,
    row: dict[str, str],
    *,
    error_accumulator: list[str] | None = None,
) -> list[URIRef]:
    rv = []
    for f in _pop_split(row, "Discipline"):
        if f.startswith("https://w3id.org/kim/hochschulfaechersystematik/"):
            rv.append(URIRef(f))
        else:
            _log(file_name, idx, f"invalid discipline: {f}", error_accumulator=error_accumulator)
    return rv


def _process_formats(row: dict[str, str]) -> list[str]:
    return [f.lstrip(".").upper() for f in _pop_split(row, "FileFormat")]


def _process_languages(row: dict[str, str]) -> list[ISO639_3]:
    return cleanup_languages(_pop_split(row, "Language"))


def _process_proficiency_levels(row: dict[str, str]) -> list[URIRef]:
    return [PROFICIENCY_LEVELS[x.lower()] for x in _pop_split(row, "ProficiencyLevel")]


def _process_authors(
    file_name: str,
    idx: int,
    row: dict[str, str],
    *,
    error_accumulator: list[str] | None = None,
) -> list[AuthorDIF13 | OrganizationDIF13]:
    return [
        author
        for s in _pop_split(row, "Authors")
        if (author := _process_author(file_name, idx, s, error_accumulator=error_accumulator))
        is not None
    ]


def _process_author(
    file_name: str,
    idx: int,
    s: str,
    *,
    error_accumulator: list[str] | None = None,
) -> AuthorDIF13 | OrganizationDIF13 | None:
    if not s or s.lower() == "n/a":
        return None

    if "{" not in s:
        # assume whole thing is a name
        family_name, _, given_name = (x.strip() for x in s.rpartition(","))
        return AuthorDIF13(given_name=given_name, family_name=family_name)

    name, _, ids = s.partition(" : ")
    url = ids.lstrip("{").rstrip("}")
    if url.startswith("organization"):
        url = url.removeprefix("organization").strip()
        if url.startswith(WIKIDATA_URI_PREFIX):
            return OrganizationDIF13(name=name, wikidata=url.removeprefix(WIKIDATA_URI_PREFIX))
        elif url.startswith(ROR_URI_PREFIX):
            return OrganizationDIF13(name=name, ror=url)
        elif not url:
            return OrganizationDIF13(name=name)
        else:
            pass
    elif url.startswith(ORCID_URI_PREFIX):
        orcid = url.removeprefix(ORCID_URI_PREFIX)
        if not ORCID_RE.fullmatch(orcid):
            _log(file_name, idx, f"invalid ORCID: {orcid}", error_accumulator=error_accumulator)
            return None
        family_name, _, given_name = (x.strip() for x in name.rpartition(","))
        return AuthorDIF13(
            given_name=given_name, family_name=family_name, orcid=ORCID_URI_PREFIX + orcid
        )
    elif url.startswith(ROR_URI_PREFIX):
        return OrganizationDIF13(name=name, ror=url.removeprefix(ROR_URI_PREFIX))

    _log(file_name, idx, f"failed to parse author: {s}", error_accumulator=error_accumulator)
    return None


def _process_license(row: dict[str, str]) -> str | URIRef | None:
    identifier = row.pop("License").strip()
    if identifier == "proprietary":
        return PROPRIETARY_LICENSE
    return SPDX_LICENSE[identifier]


def _process_target_groups(
    file_name: str,
    line: int,
    row: dict[str, str],
    *,
    error_accumulator: list[str] | None = None,
) -> list[URIRef]:
    rv = []
    for g in _pop_split(row, "TargetGroup"):
        if g.lower() in TARGET_GROUPS:
            rv.append(TARGET_GROUPS[g.lower()])
        else:
            _log(
                file_name,
                line,
                f"unable to lookup target group: {g}",
                error_accumulator=error_accumulator,
            )
    return rv


def _process_size(row: dict[str, str]) -> str | None:
    size = row.pop("Size")
    if not size:
        return None
    if size.endswith(" MB"):
        # FIXME should not be like this
        return size
    return f"{size} MB"


def _process_learning_resource_types(
    file_name: str,
    line: int,
    row: dict[str, str],
    *,
    error_accumulator: list[str] | None = None,
) -> list[URIRef]:
    rv = []
    for x in _pop_split(row, "LearningResourceType"):
        x = x.lower()
        if x.startswith("https://w3id.org/kim/hcrt/"):
            rv.append(URIRef(x))
        elif x in LEARNING_RESOURCE_TYPES:
            rv.append(LEARNING_RESOURCE_TYPES[x])
        else:
            _log(
                file_name,
                line,
                f"unable to lookup learning resource type: {x}",
                error_accumulator=error_accumulator,
            )
    return rv


def _process_media_types(
    file_name: str,
    line: int,
    row: dict[str, str],
    *,
    error_accumulator: list[str] | None = None,
) -> list[URIRef]:
    rv = []
    for g in _pop_split(row, "MediaType"):
        if g in MEDIA_TYPES:
            rv.append(MEDIA_TYPES[g])
        else:
            _log(
                file_name,
                line,
                f"unable to lookup media type: {g}",
                error_accumulator=error_accumulator,
            )
    return rv


MISSING_COMMUNITIES: Counter[str] = Counter()


def _process_communities(
    file_name: str,
    line: int,
    row: dict[str, str],
    *,
    error_accumulator: list[str] | None = None,
) -> tuple[list[URIRef], list[URIRef]]:
    supporting, recommending = [], []
    for community in _pop_split(row, "Community"):
        match = COMMUNITY_RELATION_RE.search(community)
        if not match:
            _log(
                file_name,
                line,
                f'could not match regex for community "{community}"',
                error_accumulator=error_accumulator,
            )
            continue

        name = match.group("name").strip()
        relation = match.group("relation")

        community_uuid = LOOKUP_DICT_COMMUNITIES.get(name, None)
        if not community_uuid:
            if not MISSING_COMMUNITIES[name]:
                _log(
                    file_name,
                    line,
                    f"unknown community: {name}",
                    error_accumulator=error_accumulator,
                )
            MISSING_COMMUNITIES[name] += 1
            continue

        community_uriref = DALIA_COMMUNITY[community_uuid]

        for relation_char in relation:
            match relation_char:
                case "S":
                    supporting.append(community_uriref)
                case "R":
                    recommending.append(community_uriref)

    return supporting, recommending


def _process_related_works(
    file_name: str,
    line: int,
    row: dict[str, str],
    *,
    error_accumulator: list[str] | None = None,
) -> list[PredicateObject[RDFResource]]:
    related_works = row.pop("RelatedWork")

    if not related_works.strip():
        return []

    rv: list[PredicateObject[RDFResource]] = []
    for related_work in related_works.split(DELIMITER):
        if not (related_work := related_work.strip()):
            raise Exception("Empty related work")

        related_work_substrings = related_work.split(":", maxsplit=1)

        relation = related_work_substrings[0].strip()
        relation_uriref = RELATED_WORKS_RELATIONS.get(relation, None)
        if not relation_uriref:
            _log(
                file_name,
                line,
                f'unknown related work relation "{relation}"',
                error_accumulator=error_accumulator,
            )
            continue

        if len(related_work_substrings) < 2 or not (link := related_work_substrings[1].strip()):
            _log(
                file_name,
                line,
                f"related work is missing link: `{related_work}`",
                error_accumulator=error_accumulator,
            )
            continue

        rv.append(PredicateObject(predicate=relation_uriref, object=link))
    return rv



"""RDF utilities for DIF v1.3."""

from functools import lru_cache

import pystow
import rdflib
from rdflib import XSD, Literal, URIRef
from rdflib.plugins.sparql import prepareQuery

__all__ = [
    "check_discipline_exists",
    "check_resource_type_exists",
    "get_discipline_graph",
    "get_language_graph",
    "get_language_uriref",
    "get_license_uriref",
    "get_licenses_graph",
    "get_resource_type_graph",
]

HOCHSCHULFAECHERSYSTEMATIK_TTL = "https://github.com/dini-ag-kim/hochschulfaechersystematik/raw/refs/tags/v2024-02-08/hochschulfaechersystematik.ttl"
HFS_EXISTS_QUERY = prepareQuery("""\
    PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
    ASK { ?discipline a skos:Concept . }
""")


@lru_cache(1)
def get_discipline_graph() -> rdflib.Graph:
    """Get the disciplines graph from DINI-KIM's Hochschulfaechersystematik (HSFS)."""
    return pystow.ensure_rdf("dalia", url=HOCHSCHULFAECHERSYSTEMATIK_TTL)


@lru_cache
def check_discipline_exists(discipline_uriref: URIRef) -> bool:
    result = get_discipline_graph().query(
        HFS_EXISTS_QUERY,
        initBindings={"discipline": discipline_uriref},
    )
    if result.askAnswer is None:
        raise RuntimeError
    return result.askAnswer


LICENSES_TTL = (
    "https://github.com/spdx/license-list-data/raw/refs/tags/v3.25.0/rdfturtle/licenses.ttl"
)

GET_LICENSE_URI_FROM_SPDX_QUERY = prepareQuery(
    """\
    SELECT ?license
    WHERE { ?license spdx:licenseId ?identifier }
    """,
    initNs={"spdx": "http://spdx.org/rdf/terms#"},
)


@lru_cache(1)
def get_licenses_graph() -> rdflib.Graph:
    """Get a licenses graph from SPDX."""
    graph = pystow.ensure_rdf("dalia", url=LICENSES_TTL)
    graph.bind("spdx", "http://spdx.org/rdf/terms#")
    return graph


@lru_cache
def get_license_uriref(identifier: str) -> URIRef | None:
    results = get_licenses_graph().query(
        GET_LICENSE_URI_FROM_SPDX_QUERY, initBindings={"identifier": Literal(identifier)}
    )

    if not results:
        return None

    first_result = next(results.__iter__())
    return first_result.license  # type:ignore[union-attr,return-value]


LEXVO_RDF = "http://www.lexvo.org/resources/lexvo_2013-02-09.rdf.gz"
LANGUAGE_URI_QUERY = prepareQuery("""
    PREFIX lexvo: <http://lexvo.org/ontology#>

    SELECT ?language_uri
    WHERE {
        ?language_uri lexvo:iso6392BCode|lexvo:iso6392TCode|lexvo:iso639P1Code|lexvo:iso639P3PCode ?language .
    }
""")


@lru_cache(1)
def get_language_graph() -> rdflib.Graph:
    """Get the 3-letter language code graph."""
    graph = pystow.ensure_rdf("dalia", url=LEXVO_RDF, parse_kwargs={"format": "xml"})
    graph.bind("lexvo", "http://lexvo.org/ontology#")
    return graph


@lru_cache
def get_language_uriref(language: str) -> URIRef | None:
    results = get_language_graph().query(
        LANGUAGE_URI_QUERY, initBindings={"language": Literal(language, datatype=XSD.string)}
    )

    if not results:
        return None

    first_result = next(results.__iter__())
    return first_result.language_uri  # type:ignore[union-attr,return-value]


HCRT_TTL = "https://raw.githubusercontent.com/dini-ag-kim/hcrt/3fa0effce8b07ece585c1564f047cea18eec4cad/hcrt.ttl"
HCRT_TERM_EXISTS_QUERY = prepareQuery("""\
    PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
    ASK { ?term a skos:Concept . }
""")


@lru_cache(1)
def get_resource_type_graph() -> rdflib.Graph:
    """Get the learning resource type graph from DINI-KIM's Hochschulcampus Ressourcentypen (HCRT) graph."""
    return pystow.ensure_rdf("dalia", url=HCRT_TTL)


@lru_cache
def check_resource_type_exists(hcrt_term: URIRef) -> bool:
    """Check if the resource type exists in DINI-KIM's HCRT resource."""
    result = get_resource_type_graph().query(
        HCRT_TERM_EXISTS_QUERY, initBindings={"term": hcrt_term}
    )
    if result.askAnswer is None:
        raise RuntimeError
    return result.askAnswer



"""A concrete implementation of the DALIA Interchange Format (DIF) v1.3."""

import datetime
from typing import Annotated, ClassVar

import rdflib
from pydantic import UUID4, Field
from pydantic_extra_types.language_code import ISO639_3
from pydantic_metamodel.api import (
    Addable,
    IsPredicateObject,
    PredicateAnnotation,
    PredicateObject,
    RDFInstanceBaseModel,
    RDFResource,
    WithPredicate,
    WithPredicateNamespace,
    Year,
)
from rdflib import BNode, Node, URIRef

from . import constants
from .predicates import (
    AUTHOR_PREDICATE,
    AUTHOR_UNORDERED_PREDICATE,
    DATE_PUBLISHED_PREDICATE,
    DESCRIPTION_PREDICATE,
    DISCIPLINE_PREDICATE,
    EDUCATIONAL_RESOURCE_CLASS,
    FILE_FORMAT_PREDICATE,
    FILE_SIZE_PREDICATE,
    KEYWORDS_PREDICATE,
    LANGUAGE_PREDICATE,
    LEARNING_RESOURCE_TYPE_PREDICATE,
    LICENSE_PREDICATE,
    LINK_PREDICATE,
    MEDIA_TYPES_PREDICATE,
    ORGANIZATION_CLASS,
    ORGANIZATION_NAME_PREDICATE,
    ORGANIZATION_ROR_PREDICATE,
    ORGANIZATION_WIKIDATA_PREDICATE,
    PERSON_CLASS,
    PERSON_FAMILY_NAME_PREDICATE,
    PERSON_GIVEN_NAME_PREDICATE,
    PERSON_ORCID_PREDICATE,
    PROFICIENCY_LEVEL_PREDICATE,
    RECOMMENDING_COMMUNITY_PRED,
    SUBTITLE_PREDICATE,
    SUPPORTING_COMMUNITY_PRED,
    TARGET_GROUP_PREDICATE,
    TITLE_PREDICATE,
    VERSION_PREDICATE,
)
from .utils import create_rdf_collection
from ..namespace import DALIA_OER, WIKIDATA
from ..namespace import ISO639_3 as ISO639_3_NS

__all__ = [
    "AuthorDIF13",
    "EducationalResourceDIF13",
    "OrganizationDIF13",
]


class AuthorDIF13(RDFInstanceBaseModel):
    """Represents an author in DIF v1.3."""

    rdf_type: ClassVar[URIRef] = PERSON_CLASS

    family_name: Annotated[str, WithPredicate(PERSON_FAMILY_NAME_PREDICATE)]
    given_name: Annotated[str, WithPredicate(PERSON_GIVEN_NAME_PREDICATE)]
    orcid: Annotated[str | None, WithPredicate(PERSON_ORCID_PREDICATE)] = None

    @property
    def name(self) -> str:
        """Get the full name."""
        return f"{self.given_name} {self.family_name}"

    def get_node(self) -> Node:
        """Get a blank node for the author."""
        return BNode()


class OrganizationDIF13(RDFInstanceBaseModel):
    """Represents an organization in DIF v1.3."""

    rdf_type: ClassVar[URIRef] = ORGANIZATION_CLASS

    name: Annotated[str, WithPredicate(ORGANIZATION_NAME_PREDICATE)]
    ror: Annotated[str | None, WithPredicate(ORGANIZATION_ROR_PREDICATE)] = None
    wikidata: Annotated[
        str | None, WithPredicateNamespace(ORGANIZATION_WIKIDATA_PREDICATE, WIKIDATA)
    ] = None

    def get_node(self) -> Node:
        """Get a blank node for the organization."""
        return BNode()


class AuthorAnnotation(PredicateAnnotation):
    """A custom annotation for serializing the author list."""

    def add_to_graph(self, graph: rdflib.Graph, node: Node, value: Addable) -> None:
        """Add to the graph."""
        if not isinstance(value, list):
            raise TypeError

        # the type ignores are fine because we know we're an AddableBase
        author_nodes = [self._handle_object(graph, s) for s in value]  # type:ignore[arg-type]
        if authors_list := create_rdf_collection(graph, author_nodes):
            graph.add((node, AUTHOR_PREDICATE, authors_list))

        # do this a second time to assign new blank nodes
        # (yes this is dumb, but it's for backwards compatibility)
        author_nodes = [self._handle_object(graph, s) for s in value]  # type:ignore[arg-type]
        for author in author_nodes:
            graph.add((node, AUTHOR_UNORDERED_PREDICATE, author))


class EducationalResourceDIF13(RDFInstanceBaseModel):
    """Represents an educational resource in DIF v1.3."""

    rdf_type: ClassVar[URIRef] = EDUCATIONAL_RESOURCE_CLASS

    uuid: UUID4
    title: Annotated[str, WithPredicate(TITLE_PREDICATE)] = Field(
        ..., description=constants.DIF_HEADER_TITLE.__doc__
    )
    subtitle: Annotated[str | None, WithPredicate(SUBTITLE_PREDICATE)] = None
    # TODO what about making the list ordered?
    authors: Annotated[list[AuthorDIF13 | OrganizationDIF13], AuthorAnnotation()] = Field(
        default_factory=list, description=constants.DIF_HEADER_AUTHORS.__doc__
    )
    license: Annotated[RDFResource | None, WithPredicate(LICENSE_PREDICATE)] = Field(
        None, description=constants.DIF_HEADER_LICENSE.__doc__
    )
    links: Annotated[list[RDFResource], WithPredicate(LINK_PREDICATE)] = Field(
        default_factory=list, description=constants.DIF_HEADER_LINK.__doc__
    )
    supporting_communities: Annotated[
        list[RDFResource], WithPredicate(SUPPORTING_COMMUNITY_PRED)
    ] = Field(default_factory=list, description=constants.DIF_HEADER_COMMUNITY.__doc__)
    recommending_communities: Annotated[
        list[RDFResource], WithPredicate(RECOMMENDING_COMMUNITY_PRED)
    ] = Field(default_factory=list, description=constants.DIF_HEADER_COMMUNITY.__doc__)
    description: Annotated[str | None, WithPredicate(DESCRIPTION_PREDICATE)] = Field(
        None, description=constants.DIF_HEADER_DESCRIPTION.__doc__
    )
    disciplines: Annotated[list[RDFResource] | None, WithPredicate(DISCIPLINE_PREDICATE)] = Field(
        default_factory=list, description=constants.DIF_HEADER_DISCIPLINE.__doc__
    )
    file_formats: Annotated[list[str] | None, WithPredicate(FILE_FORMAT_PREDICATE)] = Field(
        default_factory=list
    )
    keywords: Annotated[list[str], WithPredicate(KEYWORDS_PREDICATE)] = Field(default_factory=list)
    languages: Annotated[
        list[ISO639_3], WithPredicateNamespace(LANGUAGE_PREDICATE, ISO639_3_NS)
    ] = Field(default_factory=list)
    learning_resource_types: Annotated[
        list[RDFResource] | None, WithPredicate(LEARNING_RESOURCE_TYPE_PREDICATE)
    ] = Field(default_factory=list)
    media_types: Annotated[list[RDFResource], WithPredicate(MEDIA_TYPES_PREDICATE)] = Field(
        default_factory=list
    )
    proficiency_levels: Annotated[list[RDFResource], WithPredicate(PROFICIENCY_LEVEL_PREDICATE)] = (
        Field(default_factory=list)
    )
    publication_date: Annotated[
        Year | datetime.date | datetime.datetime | None, WithPredicate(DATE_PUBLISHED_PREDICATE)
    ] = None
    target_groups: Annotated[list[RDFResource] | None, WithPredicate(TARGET_GROUP_PREDICATE)] = (
        Field(default_factory=list)
    )
    related_works: Annotated[list[PredicateObject[RDFResource]], IsPredicateObject()] = Field(
        default_factory=list
    )
    file_size: Annotated[str | None, WithPredicate(FILE_SIZE_PREDICATE)] = None
    version: Annotated[str | None, WithPredicate(VERSION_PREDICATE)] = None

    def get_node(self) -> URIRef:
        """Get the learning resource URI."""
        return DALIA_OER[str(self.uuid)]



"""Utilities for DIF v1.3."""

from rdflib import RDF, BNode, Graph, Node

__all__ = [
    "create_rdf_collection",
]


def create_rdf_collection(g: Graph, elements: list[Node]) -> Node | None:
    """Create an RDF collection and return the first node, or None if the list is empty."""
    # Note: RDFLib's Collection class implementation seems to be broken. Thus, we will
    # care for the low-level triple encoding of the linked list.

    if not elements:
        return None

    list_first_node = head = BNode()
    for node in elements[0:-1]:
        g.add((head, RDF.first, node))
        rest = BNode()
        g.add((head, RDF.rest, rest))
        head = rest
    g.add((head, RDF.first, elements[-1]))
    g.add((head, RDF.rest, RDF.nil))

    return list_first_node



"""Namespaces for DALIA DIF."""

import curies
import rdflib
from rdflib import DCTERMS, RDF, SDO, Graph, Namespace

from .bibframe_lite_relation import BIBFRAME
from .bibo import BIBO
from .citedcat import CITEDCAT
from .educor import EDUCOR
from .fabio import FABIO
from .hcrt import HCRT
from .metadata4ing import METADATA4ING
from .modalia import MODALIA
from .rec import REC

__all__ = [
    "BIBFRAME",
    "BIBO",
    "CITEDCAT",
    "CONVERTER",
    "DALIA_COMMUNITY",
    "DALIA_OER",
    "DOI",
    "EDUCOR",
    "FABIO",
    "HCRT",
    "HSFS",
    "ISO639_3",
    "METADATA4ING",
    "MODALIA",
    "NAMESPACE_PREFIXES",
    "OCCO",
    "ORCID",
    "REC",
    "ROR",
    "SPDX_LICENSE",
    "T4FS",
    "WIKIDATA",
    "YOUTUBE_PLAYLIST",
    "YOUTUBE_VIDEO",
    "ZENODO_RECORD",
    "bind",
    "get_base_graph",
]

DALIA_COMMUNITY = Namespace("https://id.dalia.education/community/")
DALIA_OER = Namespace("https://id.dalia.education/learning-resource/")
ISO639_3 = Namespace("http://lexvo.org/id/iso639-3/")
HSFS = Namespace("https://w3id.org/kim/hochschulfaechersystematik/")
YOUTUBE_VIDEO = Namespace("https://www.youtube.com/watch?v=")
YOUTUBE_PLAYLIST = Namespace("https://www.youtube.com/playlist?list=")
SPDX_LICENSE = Namespace("http://spdx.org/licenses/")
WIKIDATA = Namespace("http://www.wikidata.org/entity/")
ZENODO_RECORD = Namespace("https://doi.org/10.5281/zenodo.")
OCCO = Namespace("http://purl.obolibrary.org/obo/OCCO_")
T4FS = Namespace("http://purl.obolibrary.org/obo/T4FS_")
ORCID = Namespace("https://orcid.org/")
ROR = Namespace("https://ror.org/")
DOI = Namespace("https://doi.org/")


NAMESPACE_PREFIXES: dict[str, Namespace] = {
    "schema": SDO._NS,
    "ror": ROR,
    "doi": DOI,
    "dcterms": DCTERMS._NS,
    "bflr": BIBFRAME,
    "rdf": RDF._NS,
    "bibo": BIBO,
    "citedcat": CITEDCAT,
    "dalia-community": DALIA_COMMUNITY,
    "dalia.oer": DALIA_OER,
    "educor": EDUCOR,
    "fabio": FABIO,
    "kim.hcrt": HCRT,
    "m4i": METADATA4ING,
    "modalia": MODALIA,
    "rec": REC,
    "iso639": ISO639_3,
    "hsfs": HSFS,
    "youtube.video": YOUTUBE_VIDEO,
    "youtube.playlist": YOUTUBE_PLAYLIST,
    "spdx.license": SPDX_LICENSE,
    "wikidata": WIKIDATA,
    "zenodo": ZENODO_RECORD,
    "occo": OCCO,
    "t4fs": T4FS,
}

CONVERTER = curies.Converter.from_prefix_map({k: str(v) for k, v in NAMESPACE_PREFIXES.items()})


def bind(graph: rdflib.Graph) -> None:
    """Add default namespaces to a graph."""
    for curie_prefix, uri_prefix in CONVERTER.bimap.items():
        graph.bind(curie_prefix, Namespace(uri_prefix))


def get_base_graph() -> Graph:
    """Get a graph with namespaces already bound."""
    graph = Graph()
    bind(graph)
    return graph



# mypy: disable-error-code="misc"

"""Summarize from RDF."""

from collections import Counter, defaultdict
from pathlib import Path
from textwrap import dedent
from typing import Any

import click
import matplotlib.axes
import matplotlib.pyplot as plt
import rdflib
import seaborn as sns

from dalia_dif.dif13.community import get_community_labels
from dalia_dif.dif13.legacy.rdf import get_discipline_graph
from dalia_dif.dif13.predicates import RECOMMENDING_COMMUNITY_PRED, SUPPORTING_COMMUNITY_PRED
from dalia_dif.namespace import CONVERTER

__all__ = [
    "export_chart",
]

MISSING = "missing"

COUNT_OERS_SPARQL = dedent("""\
    SELECT DISTINCT (COUNT(?s) as ?count)
    WHERE {
        ?s a educor:EducationalResource .
    }
    GROUP BY ?S
""")


def count_oers(graph: rdflib.Graph) -> int:
    """Count OERs."""
    res = graph.query(COUNT_OERS_SPARQL)
    rv = int(str(next(iter(res))[0]))  # type:ignore[index]
    return rv


COUNT_LANGUAGES_SPARQL = dedent("""\
    SELECT ?s ?o
    WHERE {
        ?s a educor:EducationalResource .
        ?s dcterms:language ?o .
    }
""")


def count_languages(graph: rdflib.Graph, *, upper: int = 4) -> tuple[Counter[str], Counter[str]]:
    """Count languages."""
    res = list(graph.query(COUNT_LANGUAGES_SPARQL))
    if len(res) == 0:
        raise ValueError(f"query returned no results:\n{COUNT_LANGUAGES_SPARQL}")
    dd = defaultdict(list)
    for uuid, lang in res:
        dd[uuid].append(str(lang).removeprefix("http://lexvo.org/id/iso639-3/"))
    combine: Counter[str] = Counter(
        ", ".join(sorted(v)) if len(v) < upper else f"eng + {len(v) - 1}" if "eng" in v else len(v)
        for v in dd.values()
    )
    single: Counter[str] = Counter(s for v in dd.values() for s in v)
    return combine, single


COUNT_LICENSES_SPARQL = dedent("""\
    SELECT ?o
    WHERE {
        ?s a educor:EducationalResource .
        ?s dcterms:license ?o .
    }
""")


def count_licenses(graph: rdflib.Graph) -> Counter[str]:
    """Count licenses."""
    res = list(graph.query(COUNT_LICENSES_SPARQL))
    if len(res) == 0:
        raise ValueError(f"query returned no results:\n{COUNT_LICENSES_SPARQL}")
    return Counter(
        CONVERTER.compress(license_uri, strict=False, passthrough=True)
        .removeprefix("spdx:")
        .removesuffix("-3.0")
        .removesuffix("-4.0")
        .removesuffix("-2.0")
        .removesuffix("-1.0")
        .replace("modalia:ProprietaryLicense", "proprietary")
        for (license_uri,) in res
    )


COUNT_FILE_EXTENSIONS_SPARQL = dedent("""\
    SELECT ?s ?o
    WHERE {
        ?s a educor:EducationalResource .
        OPTIONAL { ?s dcterms:format ?o . }
    }
""")


def count_file_extensions(graph: rdflib.Graph) -> Counter[str]:
    """Count file extensions."""
    res = list(graph.query(COUNT_FILE_EXTENSIONS_SPARQL))
    if len(res) == 0:
        raise ValueError(f"query returned no results:\n{COUNT_FILE_EXTENSIONS_SPARQL}")
    return Counter(str(format_str) if format_str else MISSING for _, format_str in res)


MEDIA_TYPE_LABELS = {
    "schema:VideoObject": "video",
    "schema:Text": "text",
    "schema:PresentationDigitalDocument": "presentation",
    "schema:AudioObject": "audio",
    "modalia:Multipart": "multipart",
    "schema:ImageObject": "image",
    "modalia:Code": "code",
}

COUNT_MEDIA_TYPE_SPARQL = dedent("""\
SELECT ?o
WHERE {
    ?s a educor:EducationalResource .
    ?s modalia:hasMediaType ?o .
}
""")


def count_media_types(graph: rdflib.Graph) -> Counter[str]:
    """Count media types."""
    res = list(graph.query(COUNT_MEDIA_TYPE_SPARQL))
    if len(res) == 0:
        raise ValueError(f"query returned no results:\n{COUNT_MEDIA_TYPE_SPARQL}")
    return Counter(
        MEDIA_TYPE_LABELS[CONVERTER.compress(media_type, strict=True)].title()
        if media_type
        else MISSING
        for (media_type,) in res
    )


COUNT_TARGET_GROUPS_SPARQL = dedent("""\
    SELECT ?o
    WHERE {
        ?s a educor:EducationalResource .
        ?s modalia:hasTargetGroup ?o .
    }
""")


def count_target_groups(graph: rdflib.Graph, total: int) -> Counter[str]:
    """Count target groups."""
    res = list(graph.query(COUNT_TARGET_GROUPS_SPARQL))
    if len(res) == 0:
        raise ValueError(f"query returned no results:\n{COUNT_TARGET_GROUPS_SPARQL}")
    rv = Counter(
        _remap_target_group(CONVERTER.parse_uri(target_group, strict=True).identifier)
        if target_group
        else MISSING
        for (target_group,) in res
    )

    most, count = rv.most_common(1)[0]
    click.echo(f"The largest target group was {most} ({count:,}/{total:,}; {count / total:.1%})")
    return rv


TARGET_GROUP_RENAMES = {
    "PhDStudent": "PhD Student",
    "MastersStudent": "Master Student",
    "BachelorStudent": "Bachelor Student",
    "StudentSchool": "Bachelor Student",
    "DataSteward": "Data Steward",
    "TeacherHighEducation": "Teacher (Higher Ed.)",
    "TeacherSchool": "Teacher (Lower Ed.)",
    "ContentProvider": "Content Provider",
}


def _remap_target_group(x: str) -> str:
    return TARGET_GROUP_RENAMES.get(x, x)


LRT_MAPPING = {
    "PodcastSeries": "podcast",
    "drill_and_practice": "drill and practice",
    "bestpractices": "best practices",
    "Bestpractices": "best practices",
    "codenotebook": "code notebook",
    "Codenotebook": "code notebook",
    "softwaresourcecode": "code",
    "Softwaresourcecode": "code",
}

COUNT_LEARNING_TYPE_RESOURCES_SPARQL = dedent("""\
    SELECT ?o
    WHERE {
        ?s a educor:EducationalResource .
        ?s modalia:hasLearningType ?o .
    }
""")


def count_learning_resource_type(graph: rdflib.Graph) -> Counter[str]:
    """Count learning resource type."""
    res = list(graph.query(COUNT_LEARNING_TYPE_RESOURCES_SPARQL))
    if len(res) == 0:
        raise ValueError(f"query returned no results:\n{COUNT_LEARNING_TYPE_RESOURCES_SPARQL}")
    return Counter(
        _remap_lrt(CONVERTER.parse_uri(learning_resource_type, strict=True).identifier)
        if learning_resource_type
        else MISSING
        for (learning_resource_type,) in res
    )


def _remap_lrt(x: str) -> str:
    return LRT_MAPPING.get(x, x).replace("_", " ").title()


DISCIPLINES_RENAMES = {
    "Cultural Studies in the narrower sense": "Cultural Studies",
    "Archival and Documentation Science": "Archival and Documentation",
    "Human Medicine / Health Sciences": "Health Sciences",
    "Information and Library Sciences": "Library Sciences",
}

GET_DISCIPLINE_LABEL_SPARQL = dedent("""\
    SELECT ?s ?label WHERE {
        ?s <http://www.w3.org/2004/02/skos/core#prefLabel> ?label .
    }
""")


def get_discipline_names() -> dict[str, str]:
    """Get discipline names (LUID to name map for HSFS)."""
    dd: defaultdict[str, list[tuple[str | None, str]]] = defaultdict(list)
    for iri, label in get_discipline_graph().query(GET_DISCIPLINE_LABEL_SPARQL):
        if not isinstance(label, rdflib.Literal):
            raise ValueError
        luid = iri.removeprefix("https://w3id.org/kim/hochschulfaechersystematik/")
        dd[luid].append((label._language, DISCIPLINES_RENAMES.get(label._value, label._value)))
    return {k: min(v, key=_get_best_lang)[1] for k, v in dd.items()}


def _get_best_lang(pair: tuple[str | None, Any]) -> tuple[int, str]:
    if pair[0] == "en" or pair[0] is None:
        return 0, ""
    elif pair[0] == "de":
        return 1, ""
    else:
        return 2, pair[0]


COUNT_PROFICIENCY_LEVELS_SPARQL = dedent("""\
    SELECT ?s ?o
    WHERE {
        ?s a educor:EducationalResource .
        OPTIONAL { ?s modalia:requiresProficiencyLevel ?o . }
    }
""")


def count_proficiency_level(graph: rdflib.Graph, total: int) -> Counter[str]:
    """Count proficiency level."""
    res = list(graph.query(COUNT_PROFICIENCY_LEVELS_SPARQL))
    if len(res) == 0:
        raise ValueError(f"query returned no results:\n{COUNT_PROFICIENCY_LEVELS_SPARQL}")
    rv = Counter(
        CONVERTER.parse_uri(proficiency_level, strict=True).identifier.title()
        if proficiency_level
        else MISSING
        for subj, proficiency_level in res
    )

    most, count = rv.most_common(1)[0]
    click.echo(
        f"The largest proficiency level was {most} ({count:,}/{total:,}; {count / total:.1%})"
    )
    return rv


COUNT_DISCIPLINES_SPARQL = dedent("""\
    SELECT ?o
    WHERE {
        ?s a educor:EducationalResource .
        ?s <http://purl.org/spar/fabio/hasDiscipline> ?o .
    }
""")


def count_disciplines(graph: rdflib.Graph) -> Counter[str]:
    """Count disciplines."""
    res = list(graph.query(COUNT_DISCIPLINES_SPARQL))
    if len(res) == 0:
        raise ValueError(f"query returned no results:\n{COUNT_DISCIPLINES_SPARQL}")
    names = get_discipline_names()
    rv = Counter(
        names[CONVERTER.parse_uri(discipline, strict=True).identifier] for (discipline,) in res
    )
    frv: Counter[str] = Counter()
    for k, v in rv.most_common():
        if v > 1:
            frv[k] = v
        else:
            frv["Other"] += v
    return frv


COUNT_COMMUNITIES_SPARQL = dedent(f"""\
    SELECT ?o
    WHERE {{
        ?s a educor:EducationalResource .
        ?s ?p ?o
        VALUES ?p {{ <{SUPPORTING_COMMUNITY_PRED}> <{RECOMMENDING_COMMUNITY_PRED}> }}
    }}
""")


def count_communities(graph: rdflib.Graph) -> Counter[str]:
    """Count communities."""
    community_labels = get_community_labels()
    res = list(graph.query(COUNT_COMMUNITIES_SPARQL))
    if len(res) == 0:
        raise ValueError(f"query returned no results:\n{COUNT_COMMUNITIES_SPARQL}")
    # TODO this should be in the graph!
    rv = Counter(
        community_labels[str(community).removeprefix("https://id.dalia.education/community/")]
        for (community,) in res
    )
    for k, v in rv.most_common():
        if v < 3:
            rv["Other"] += v
            del rv[k]
    return rv


def barplot_counter(
    counter: Counter[str],
    *,
    ax: matplotlib.axes.Axes | None = None,
    title: str | None = None,
    log: bool = True,
) -> matplotlib.axes.Axes:
    """Plot a counter."""
    categories, counts = zip(*counter.most_common(), strict=False)
    rv = sns.barplot(y=categories, x=counts, ax=ax)
    if title:
        rv.set_title(title)
    if log:
        rv.set_xscale("log")
    return rv


def export_chart(graph: rdflib.Graph, paths: Path | list[Path]) -> None:
    """Export the chart."""
    n_oers = count_oers(graph)
    click.secho(f"DALIA has {n_oers:,} OERs")

    _fig, axes = plt.subplots(3, 3, figsize=(15, 17))

    _combine_language_counter, single_language_counter = count_languages(graph)
    barplot_counter(single_language_counter, ax=axes[0][0], title="Language Occurrence")
    barplot_counter(count_licenses(graph), ax=axes[0][1], title="Licenses")
    barplot_counter(count_file_extensions(graph), ax=axes[0][2], title="File Extensions")
    barplot_counter(count_media_types(graph), ax=axes[1][0], title="Media Types")
    barplot_counter(
        count_proficiency_level(graph, n_oers), ax=axes[1][1], title="Required Proficiency Level"
    )
    barplot_counter(count_disciplines(graph), ax=axes[1][2], title="Disciplines")
    barplot_counter(count_communities(graph), ax=axes[2][0], title="Community")
    barplot_counter(count_target_groups(graph, n_oers), ax=axes[2][1], title="Target Groups")
    barplot_counter(
        count_learning_resource_type(graph), ax=axes[2][2], title="Learning Resource Type"
    )

    plt.tight_layout()
    if not isinstance(paths, list):
        paths = [paths]
    for path in paths:
        plt.savefig(path, dpi=300)
    plt.close()



"""Terms from FaBiO, the FRBR-aligned Bibliographic Ontology."""

from rdflib import Namespace

__all__ = [
    "FABIO",
]
FABIO = Namespace("http://purl.org/spar/fabio/")

# Properties
# https://sparontologies.github.io/fabio/current/fabio.html#d4e205
hasDiscipline = FABIO["hasDiscipline"]
# https://sparontologies.github.io/fabio/current/fabio.html#d4e1689
hasSubtitle = FABIO["hasSubtitle"]



"""Terms from the Bibliographic Ontology (https://dcmi.github.io/bibo/)."""

from rdflib import Namespace

__all__ = [
    "BIBO",
]

BIBO = Namespace("http://purl.org/ontology/bibo/")

# Types
Article = BIBO["Article"]
Book = BIBO["Book"]
Report = BIBO["Report"]
Thesis = BIBO["Thesis"]
Webpage = BIBO["Webpage"]



"""Terms from EduCOR (https://github.com/tibonto/educor/blob/main/educor.ttl)."""

from rdflib import Namespace

__all__ = [
    "EDUCOR",
]
EDUCOR = Namespace("https://github.com/tibonto/educor#")

# Types
EducationalResource = EDUCOR["EducationalResource"]



"""Utilities."""

from pydantic_extra_types.language_code import ISO639_3, _index_by_alpha2
from tqdm import tqdm

__all__ = [
    "cleanup_languages",
]


def cleanup_languages(language_codes: str | list[str]) -> list[ISO639_3]:
    """Extract languages from a code string."""
    if isinstance(language_codes, str):
        language_codes = [language_codes]
    r = []
    for language_code in language_codes:
        if language_code == "N/A":
            continue
        if len(language_code) == 2:
            r.append(ISO639_3(_index_by_alpha2()[language_code].alpha3))
        elif len(language_code) == 3:
            r.append(ISO639_3(language_code))
        else:
            tqdm.write(f"can not handle language code: {language_code}")
    return r



"""Terms from Hochschulcampus Ressourcentypen (https://skohub.io/dini-ag-kim/hcrt/heads/master/w3id.org/kim/hcrt/scheme.html)."""

from rdflib import Namespace

__all__ = [
    "HCRT",
]
HCRT = Namespace("https://w3id.org/kim/hcrt/")

# Types
audio = HCRT["audio"]
case_study = HCRT["case_study"]
diagram = HCRT["diagram"]
educational_game = HCRT["educational_game"]
experiment = HCRT["experiment"]
image = HCRT["image"]
slide = HCRT["slide"]
text = HCRT["text"]
video = HCRT["video"]
course = HCRT["course"]



"""An implementation of DIF v1.3."""

from .model import AuthorDIF13, EducationalResourceDIF13, OrganizationDIF13
from .reader import (
    parse_dif13_row,
    read_dif13,
    read_dif13_into_rdflib,
    write_dif13_jsonl,
    write_dif13_rdf,
)

__all__ = [
    "AuthorDIF13",
    "EducationalResourceDIF13",
    "OrganizationDIF13",
    "parse_dif13_row",
    "read_dif13",
    "read_dif13_into_rdflib",
    "write_dif13_jsonl",
    "write_dif13_rdf",
]



"""Terms from Bibframe Lite + Relation (http://bibfra.me/vocab/relation/)."""

from rdflib import Namespace

__all__ = [
    "BIBFRAME",
]

BIBFRAME = Namespace("http://bibfra.me/vocab/relation/")

# Properties
supportinghost = BIBFRAME["supportinghost"]



"""A legacy implementation of DIF v1.3 from Frank Lange that parses direclty into RDF."""

from .learning_resource import parse_dif13_row_legacy, read_dif13

__all__ = [
    "parse_dif13_row_legacy",
    "read_dif13",
]



"""Command line interface for :mod:`dalia_dif`."""

import sys
from pathlib import Path

import click

__all__ = [
    "main",
]


@click.group()
def main() -> None:
    """CLI for dalia_dif."""


@main.command()
@click.option("--dif-version", type=click.Choice(["1.3"]), default="1.3")
@click.argument("location")
def validate(location: str, dif_version: str) -> None:
    """Validate a local/remote file or local folder of DIF-encoded CSVs."""
    from dalia_dif.dif13 import read_dif13

    fail = False
    p = Path(location)
    if p.exists() and p.is_dir():
        p = p.expanduser().resolve()

        click.echo(f"validating directory: {p}")
        for path in p.glob("*.csv"):
            click.secho(f"\n> {path.relative_to(p)}", fg="green")
            errors: list[str] = []
            read_dif13(path, error_accumulator=errors)
            if errors:
                fail = True
                for error in errors:
                    click.secho(error, fg="red")
        if fail:
            click.secho("validation failed", fg="red")
            sys.exit(1)

    else:
        errors = []
        read_dif13(location, error_accumulator=errors)
        if errors:
            for error in errors:
                click.secho(error, fg="red")

            click.secho("validation failed", fg="red")
            sys.exit(1)


@main.command()
@click.argument("location", type=Path)
def lint(location: Path) -> None:
    """Lint CSV files in the curation directory."""
    import uuid

    import pandas as pd

    from .dif13.constants import DIF_HEADER_ID

    count = 0
    for path in location.glob("*.csv"):
        df = pd.read_csv(path, sep=",")
        if DIF_HEADER_ID not in df.columns:
            click.secho(f"missing column {DIF_HEADER_ID} in {path}")
            new_columns = [DIF_HEADER_ID, *df.columns]
            df[DIF_HEADER_ID] = df.index.map(lambda _: str(uuid.uuid4()))
            df = df[new_columns]
            df.to_csv(path, sep=",", index=False)
        count += len(df)
    click.echo(f"There are a total of {count:,} rows")


@main.command()
@click.option("--dif-version", type=click.Choice(["1.3"]), default="1.3")
@click.option("--format")
@click.option("-o", "--output", type=Path)
@click.argument("location")
def convert(location: str, dif_version: str, format: str | None, output: Path | None) -> None:
    """Validate a DIF file."""
    from dalia_dif.dif13 import read_dif13, write_dif13_jsonl, write_dif13_rdf

    oers = read_dif13(location)

    if output is None:
        if format == "jsonl":
            write_dif13_jsonl(oers)
        else:
            write_dif13_rdf(oers)
    elif output.suffix == ".ttl":
        write_dif13_rdf(oers, path=output, format=format)
    elif output.suffix == ".jsonl":
        write_dif13_jsonl(oers, path=output)
    else:
        click.secho(f"unhandled extension: {output.suffix}. Use .ttl or .jsonl", fg="red")


if __name__ == "__main__":
    main()



"""Code for curated DALIA communities."""

import csv
from collections import Counter
from pathlib import Path

HERE = Path(__file__).parent.resolve()
COMMUNITIES_PATH = HERE / "dalia_communities.csv"


def _read_mapping() -> dict[str, str]:
    rv = {}
    with open(COMMUNITIES_PATH, newline="") as csvfile:
        for row in csv.DictReader(csvfile):
            rv[row["Title"]] = row["ID"]
            for synonym in row["Synonyms"].split("|"):
                rv[synonym] = row["ID"]
    return rv


LOOKUP_DICT_COMMUNITIES = _read_mapping()

del _read_mapping

MISSING_COMMUNITIES: Counter[str] = Counter()


def get_community_labels() -> dict[str, str]:
    """Get community labels."""
    with open(COMMUNITIES_PATH, newline="") as csvfile:
        return {
            row["ID"]: COMMUNITY_RELABELS.get(row["Title"], row["Title"])
            for row in csv.DictReader(csvfile)
        }


COMMUNITY_RELABELS = {
    "Nationale Forschungsdateninfrastruktur (NFDI)": "NFDI",
    "HeFDI - Hessische Forschungsdateninfrastrukturen": "HeFDI",
}



"""Terms from Metadata4Ing (https://w3id.org/nfdi4ing/metadata4ing)."""

from rdflib import Namespace

__all__ = [
    "METADATA4ING",
]
METADATA4ING = Namespace("http://w3id.org/nfdi4ing/metadata4ing#")

# Properties
hasRorId = METADATA4ING["hasRorId"]
orcidId = METADATA4ING["orcidId"]



"""A refactored legacy implementation parsing DIF v1.3 from Frank Lange."""

from pathlib import Path
from uuid import UUID

import click
from pystow.utils import safe_open_dict_reader
from rdflib import RDF, Graph
from tqdm import tqdm

from .authors import add_authors_to_lr
from .components import (
    add_communities_to_lr,
    add_description_to_lr,
    add_disciplines_to_lr,
    add_file_formats_to_lr,
    add_keywords_to_lr,
    add_languages_to_lr,
    add_learning_resource_types_to_lr,
    add_license_to_lr,
    add_links_to_lr,
    add_media_types_to_lr,
    add_proficiency_levels_to_lr,
    add_publication_date_to_lr,
    add_related_works_to_lr,
    add_size_to_lr,
    add_target_groups_to_lr,
    add_title_to_lr,
    add_version_to_lr,
)
from .. import constants
from ..constants import DIF_HEADER_ID
from ..predicates import EDUCATIONAL_RESOURCE_CLASS
from ...namespace import DALIA_OER, get_base_graph

__all__ = [
    "parse_dif13_row_legacy",
    "read_dif13",
]


def read_dif13(filename: str | Path, *, sep: str = ",") -> tuple[bool, Graph]:
    """Read a CSV containing DIF v1.3 encoded content.

    :param filename: The local filepath.
    :param sep: The seperator for the CSV file.

    :returns: A pair of a boolean "has error" and a RDFlib graph
    """
    has_error = False
    collection_graph = get_base_graph()

    path = Path(filename).expanduser().resolve()
    with safe_open_dict_reader(path, delimiter=sep) as reader:
        if DIF_HEADER_ID not in set(reader.fieldnames):  # type:ignore[arg-type]
            has_error = True
            tqdm.write(click.style(f"no {DIF_HEADER_ID} column found in {path}", fg="red"))
            return has_error, collection_graph

        for row_number, row in enumerate(reader, start=2):
            try:
                parse_dif13_row_legacy(collection_graph, row_number, _trim_row(row), path=path)
            except Exception as e:
                tqdm.write(click.style(f"{e!s}", fg="red"))
                if isinstance(e, KeyError):
                    raise ValueError("This could be due to an incorrect column name.") from None
                has_error = True

    return has_error, collection_graph


def _trim_row(row: dict[str, str]) -> dict[str, str]:
    return {k: v.strip() for k, v in row.items()}


def parse_dif13_row_legacy(
    g: Graph, row_number: int, row: dict[str, str], path: Path | None = None
) -> None:
    if not (lr_id := row[constants.DIF_HEADER_ID].strip()):
        tqdm.write(
            click.style(
                f"[{row_number}] Empty {constants.DIF_HEADER_ID} field - ignoring this row",
                fg="yellow",
            )
        )
        return None

    uuid = UUID(lr_id)  # Validation: may rise ValueError
    lr_node = DALIA_OER[str(uuid)]
    g.add((lr_node, RDF.type, EDUCATIONAL_RESOURCE_CLASS))

    add_authors_to_lr(g, lr_node, row[constants.DIF_HEADER_AUTHORS])
    add_license_to_lr(g, lr_node, row[constants.DIF_HEADER_LICENSE])
    add_links_to_lr(g, lr_node, row[constants.DIF_HEADER_LINK])
    add_title_to_lr(g, lr_node, row[constants.DIF_HEADER_TITLE])
    add_communities_to_lr(
        g, lr_node, row[constants.DIF_HEADER_COMMUNITY], row_number=row_number, path=path
    )
    add_description_to_lr(g, lr_node, row[constants.DIF_HEADER_DESCRIPTION])
    add_disciplines_to_lr(
        g, lr_node, row[constants.DIF_HEADER_DISCIPLINE], row_number=row_number, path=path
    )
    add_file_formats_to_lr(g, lr_node, row[constants.DIF_HEADER_FILE_FORMAT])
    add_keywords_to_lr(g, lr_node, row[constants.DIF_HEADER_KEYWORDS])
    add_languages_to_lr(g, lr_node, row[constants.DIF_HEADER_LANGUAGE])
    add_learning_resource_types_to_lr(
        g,
        lr_node,
        row[constants.DIF_HEADER_LEARNING_RESOURCE_TYPE],
        row_number=row_number,
        path=path,
    )
    add_media_types_to_lr(g, lr_node, row[constants.DIF_HEADER_MEDIA_TYPE])
    add_proficiency_levels_to_lr(g, lr_node, row[constants.DIF_HEADER_PROFICIENCY_LEVEL])
    add_publication_date_to_lr(g, lr_node, row[constants.DIF_HEADER_PUBLICATION_DATE])
    add_target_groups_to_lr(g, lr_node, row[constants.DIF_HEADER_TARGET_GROUP])
    add_related_works_to_lr(g, lr_node, row[constants.DIF_HEADER_RELATED_WORK])
    add_size_to_lr(g, lr_node, row[constants.DIF_HEADER_SIZE])
    add_version_to_lr(g, lr_node, row[constants.DIF_HEADER_VERSION])

    return None



"""A refactored legacy implementation parsing DIF v1.3 from Frank Lange."""

import re
from pathlib import Path

import click
from rdflib import XSD, Graph, Literal, Node, URIRef

from .rdf import (
    check_discipline_exists,
    check_resource_type_exists,
    get_language_uriref,
    get_license_uriref,
)
from ..community import COMMUNITIES_PATH, LOOKUP_DICT_COMMUNITIES, MISSING_COMMUNITIES
from ..constants import DIF_SEPARATOR
from ..picklists import (
    COMMUNITY_RELATIONS,
    LEARNING_RESOURCE_TYPES,
    MEDIA_TYPE_EXCEPTIONS,
    MEDIA_TYPES,
    PROFICIENCY_LEVELS,
    PROPRIETARY_LICENSE,
    RELATED_WORKS_RELATIONS,
    TARGET_GROUPS,
)
from ..predicates import (
    DATE_PUBLISHED_PREDICATE,
    DESCRIPTION_PREDICATE,
    DISCIPLINE_PREDICATE,
    FILE_FORMAT_PREDICATE,
    FILE_SIZE_PREDICATE,
    KEYWORDS_PREDICATE,
    LANGUAGE_PREDICATE,
    LEARNING_RESOURCE_TYPE_PREDICATE,
    LICENSE_PREDICATE,
    LINK_PREDICATE,
    MEDIA_TYPES_PREDICATE,
    PROFICIENCY_LEVEL_PREDICATE,
    SUBTITLE_PREDICATE,
    TARGET_GROUP_PREDICATE,
    TITLE_PREDICATE,
    VERSION_PREDICATE,
)
from ...namespace import DALIA_COMMUNITY


def add_description_to_lr(g: Graph, lr_node: Node, description: str) -> None:
    if not (description := description.strip()):
        return
    g.add((lr_node, DESCRIPTION_PREDICATE, Literal(description)))


def add_file_formats_to_lr(g: Graph, lr_node: Node, file_formats: str) -> None:
    if not file_formats.strip():
        return

    for file_format in file_formats.split(DIF_SEPARATOR):
        if not (file_format := file_format.strip()):
            raise Exception("Empty file format")

        match = re.search(r"^\.?(?P<format>\S+)$", file_format)
        if not match:
            raise Exception(f'Could not match regex for file format "{file_format}"')
        fmt = match.group("format")

        g.add((lr_node, FILE_FORMAT_PREDICATE, Literal(fmt.upper())))


def add_keywords_to_lr(g: Graph, lr_node: Node, keywords: str) -> None:
    if not keywords.strip():
        return

    for keyword in keywords.split(DIF_SEPARATOR):
        if not (keyword := keyword.strip()):
            raise Exception("Empty keyword")

        g.add((lr_node, KEYWORDS_PREDICATE, Literal(keyword)))


def add_media_types_to_lr(g: Graph, lr_node: Node, media_types: str) -> None:
    if not media_types.strip():
        return

    for media_type in media_types.split(DIF_SEPARATOR):
        if not (media_type := media_type.strip()):
            raise Exception("Empty media type")

        media_type_uriref = MEDIA_TYPES.get(media_type.lower(), None)
        if not media_type_uriref:
            raise Exception(f'Unknown media type "{media_type}"')

        g.add((lr_node, MEDIA_TYPES_PREDICATE, media_type_uriref))


def add_proficiency_levels_to_lr(g: Graph, lr_node: Node, proficiency_levels: str) -> None:
    if not proficiency_levels.strip():
        return

    for proficiency_level in proficiency_levels.split(DIF_SEPARATOR):
        if not (proficiency_level := proficiency_level.strip()):
            raise Exception("Empty proficiency level")

        proficiency_level_uriref = PROFICIENCY_LEVELS.get(proficiency_level.lower(), None)
        if not proficiency_level_uriref:
            raise Exception(f'Unknown proficiency level "{proficiency_level}"')

        g.add((lr_node, PROFICIENCY_LEVEL_PREDICATE, proficiency_level_uriref))


def add_size_to_lr(g: Graph, lr_node: Node, size: str) -> None:
    if not (size := size.strip()):
        return

    float(size.split()[0])  # Validation: may rise ValueError

    g.add(
        (
            lr_node,
            FILE_SIZE_PREDICATE,
            Literal(f"{size} MB" if bool(re.fullmatch(r"\d+(\.\d+)?", size)) else size),
        )
    )


def add_target_groups_to_lr(g: Graph, lr_node: Node, target_groups: str) -> None:
    if not target_groups.strip():
        return

    for target_group in target_groups.split(DIF_SEPARATOR):
        if not (target_group := target_group.strip()):
            raise Exception("Empty target group")

        target_group_uriref = TARGET_GROUPS.get(target_group.lower(), None)
        if not target_group_uriref:
            raise Exception(f'Unknown target group "{target_group}"')

        g.add((lr_node, TARGET_GROUP_PREDICATE, target_group_uriref))


def add_title_to_lr(g: Graph, lr_node: Node, title: str) -> None:
    if not (title := title.strip()):
        raise Exception("Empty Title field")

    title_substrings = title.split(": ", maxsplit=1)

    g.add((lr_node, TITLE_PREDICATE, Literal(title_substrings[0].strip())))

    if len(title_substrings) > 1 and (subtitle := title_substrings[1].strip()):
        g.add((lr_node, SUBTITLE_PREDICATE, Literal(subtitle)))


def add_related_works_to_lr(g: Graph, lr_node: Node, related_works: str) -> None:
    if not related_works.strip():
        return

    for related_work in related_works.split(DIF_SEPARATOR):
        if not (related_work := related_work.strip()):
            raise Exception("Empty related work")

        related_work_substrings = related_work.split(":", maxsplit=1)

        relation = related_work_substrings[0].strip()
        relation_uriref = RELATED_WORKS_RELATIONS.get(relation, None)
        if not relation_uriref:
            raise Exception(f'Unknown related work relation "{relation}"')

        if len(related_work_substrings) < 2 or not (link := related_work_substrings[1].strip()):
            raise Exception(f'Link missing in related work "{related_work}"')

        g.add((lr_node, relation_uriref, URIRef(link)))


def add_links_to_lr(g: Graph, lr_node: Node, links: str) -> None:
    if not links.strip():
        raise Exception("Empty Link field")

    for link in links.split(DIF_SEPARATOR):
        if not (link := link.strip()):
            raise Exception("Empty link")

        g.add((lr_node, LINK_PREDICATE, URIRef(link)))


def add_version_to_lr(g: Graph, lr_node: Node, version: str) -> None:
    if not (version := version.strip()):
        return

    g.add((lr_node, VERSION_PREDICATE, Literal(version)))


def add_disciplines_to_lr(
    g: Graph, lr_node: Node, disciplines: str, row_number: int = 0, path: Path | None = None
) -> None:
    if not disciplines.strip():
        return

    for discipline in disciplines.split(DIF_SEPARATOR):
        if not (discipline := discipline.strip()):
            raise Exception("Empty discipline")

        discipline_uriref = URIRef(discipline)
        if not check_discipline_exists(discipline_uriref):
            raise ValueError(
                f'[{path} line:{row_number:,}] discipline "{discipline}" does not exist in the Hochschulfaechersystematik'
            )

        g.add((lr_node, DISCIPLINE_PREDICATE, discipline_uriref))


def add_license_to_lr(g: Graph, lr_node: Node, license_: str) -> None:
    if not (license_ := license_.strip()):
        raise Exception("Empty License field")

    license_uri: URIRef | None
    if license_ == "proprietary":
        license_uri = PROPRIETARY_LICENSE
    else:
        license_uri = get_license_uriref(license_)
    if not license_uri:
        raise Exception(f'Invalid license identifier "{license_}". Please check SPDX.')

    g.add((lr_node, LICENSE_PREDICATE, license_uri))


def add_languages_to_lr(g: Graph, lr_node: Node, languages: str) -> None:
    if not languages.strip():
        return

    for language in languages.split(DIF_SEPARATOR):
        if not (language := language.strip()):
            raise Exception("Empty language")

        language_uri = get_language_uriref(language)
        if not language_uri:
            raise Exception(f'Invalid language identifier "{language}". Please check Lexvo.')

        g.add((lr_node, LANGUAGE_PREDICATE, language_uri))


def add_publication_date_to_lr(g: Graph, lr_node: Node, publication_date: str) -> None:
    if not (publication_date := publication_date.strip()):
        return
    d = _date_to_literal(publication_date)
    if d is None:
        raise ValueError(f'Invalid publication date "{publication_date}"')
    g.add((lr_node, DATE_PUBLISHED_PREDICATE, d))


def _date_to_literal(s: str) -> Literal | None:
    # regex from https://stackoverflow.com/a/22061879
    # YYYY-MM-DD
    if re.search(r"^\d{4}-(0[1-9]|1[012])-(0[1-9]|[12][0-9]|3[01])$", s):
        return Literal(s, datatype=XSD.date)
    # YYYY-MM
    if re.search(r"^\d{4}-(0[1-9]|1[012])$", s):
        return Literal(s, datatype=XSD.gYearMonth)
    # YYYY
    if re.search(r"^\d{4}$", s):
        return Literal(s, datatype=XSD.gYear)
    # DD.MM.YYYY
    if re.search(r"^(0[1-9]|[12][0-9]|3[01])\.(0[1-9]|1[012])\.\d{4}$", s):
        parts = s.split(".")
        s = f"{parts[2]}-{parts[1]}-{parts[0]}"
        return Literal(s, datatype=XSD.date)
    # MM.YYYY
    if re.search(r"^(0[1-9]|1[012])\.\d{4}$", s):
        parts = s.split(".")
        s = f"{parts[1]}-{parts[0]}"
        return Literal(s, datatype=XSD.gYearMonth)
    return None


def add_learning_resource_types_to_lr(
    g: Graph,
    lr_node: Node,
    learning_resource_types: str,
    row_number: int = 0,
    path: Path | None = None,
) -> None | list[URIRef | None]:
    if not learning_resource_types.strip():
        return None

    rv: list[URIRef | None] = []
    for learning_resource_type in learning_resource_types.split(DIF_SEPARATOR):
        if not (learning_resource_type := learning_resource_type.strip()):
            raise Exception("Empty learning resource type")

        lr_type_uriref = LEARNING_RESOURCE_TYPES.get(learning_resource_type.lower(), None)
        if not lr_type_uriref:
            # Try to find it in the HCRT vocabulary.
            hcrt_term = URIRef(learning_resource_type)
            if check_resource_type_exists(hcrt_term):
                lr_type_uriref = hcrt_term

        if not lr_type_uriref:
            raise ValueError(f'Unknown learning resource type "{learning_resource_type}"')

        if lr_type_uriref in MEDIA_TYPE_EXCEPTIONS:
            if path:
                lead = f"[{path.name} line:{row_number}]"
            else:
                lead = f"[{row_number}]"

            click.secho(
                f'{lead} rather than using "{learning_resource_type}" resource type, consider '
                f'using media type "{MEDIA_TYPE_EXCEPTIONS[lr_type_uriref]}"',
                fg="yellow",
            )
            continue

        g.add((lr_node, LEARNING_RESOURCE_TYPE_PREDICATE, lr_type_uriref))
        rv.append(lr_type_uriref)
    return rv


def add_communities_to_lr(
    g: Graph, lr_node: Node, communities: str, *, row_number: int = 0, path: Path | None = None
) -> None:
    if not communities.strip():
        return

    for community in communities.split(DIF_SEPARATOR):
        if not (community := community.strip()):
            raise Exception("Empty community")

        match = re.search(r"^(?P<name>.*)\s\((?P<relation>S|R|SR|RS)\)$", community)
        if not match:
            raise ValueError(
                f'[line:{row_number:,}] Community was incorrectly encoded "{community}". Did you remember to '
                f"include one of (S), (R), (SR), or (RS) at the end?"
            )

        name = match.group("name").strip()
        relation = match.group("relation")

        community_id = LOOKUP_DICT_COMMUNITIES.get(name, None)
        if not community_id:
            if not MISSING_COMMUNITIES[name]:
                msg = f'unknown community "{name}".\n\tSuggestion: add to {COMMUNITIES_PATH.name}'
                if path:
                    msg = f"[{path.name}] {msg}"
                raise ValueError(msg)
            MISSING_COMMUNITIES[name] += 1
            return
        community_uriref = DALIA_COMMUNITY[community_id]

        for relation_char in relation:
            relation_uriref = COMMUNITY_RELATIONS[relation_char]
            g.add((lr_node, relation_uriref, community_uriref))



"""Terms from DataCite to DCAT-AP Mapping (https://ec-jrc.github.io/datacite-to-dcat-ap/)."""

from rdflib import Namespace

__all__ = [
    "CITEDCAT",
]
CITEDCAT = Namespace("https://w3id.org/citedcat-ap/")

# Properties
isSupplementedBy = CITEDCAT["isSupplementedBy"]
isSupplementTo = CITEDCAT["isSupplementTo"]



"""Terms from the MoDalia ontology (https://git.rwth-aachen.de/dalia/dalia-ontology/-/blob/main/MoDalia.ttl)."""

from rdflib.namespace import Namespace

__all__ = [
    "MODALIA",
]

MODALIA = Namespace("https://purl.org/ontology/modalia#")

# Properties
hasLearningType = MODALIA["hasLearningType"]
hasMediaType = MODALIA["hasMediaType"]
hasTargetGroup = MODALIA["hasTargetGroup"]
isBasedOn = MODALIA["isBasedOn"]
isDuplicateOf = MODALIA["isDuplicateOf"]
isNewVersionOf = MODALIA["isNewVersionOf"]
isPartOf = MODALIA["isPartOf"]
isTranslationOf = MODALIA["isTranslationOf"]
requiresProficiencyLevel = MODALIA["requiresProficiencyLevel"]

# Types
BachelorStudent = MODALIA["BachelorStudent"]
BestPractices = MODALIA["BestPractices"]
Code = MODALIA["Code"]
CodeNotebook = MODALIA["CodeNotebook"]
ContentProvider = MODALIA["ContentProvider"]
Cookbook = MODALIA["Cookbook"]
DataSteward = MODALIA["DataSteward"]
Lecture = MODALIA["Lecture"]
MastersStudent = MODALIA["MastersStudent"]
Multipart = MODALIA["Multipart"]
PhDStudent = MODALIA["PhDStudent"]
Poster = MODALIA["Poster"]
Researcher = MODALIA["Researcher"]
StudentSchool = MODALIA["StudentSchool"]
TeacherHighEducation = MODALIA["TeacherHighEducation"]
TeacherSchool = MODALIA["TeacherSchool"]
Tutorial = MODALIA["Tutorial"]
Interview = MODALIA["Interview"]  # TODO review
Course = MODALIA["Interview"]  # TODO review

# Individuals
Beginner = MODALIA["Beginner"]
Competent = MODALIA["Competent"]
Expert = MODALIA["Expert"]
Novice = MODALIA["Novice"]
Proficient = MODALIA["Proficient"]
ProprietaryLicense = MODALIA["ProprietaryLicense"]
